import java.util.concurrent.Semaphore;
import java.util.concurrent.locks.ReentrantLock;


/*
 * Three philosophers dine together at the same table. Each philosopher
 * has their own plate at the table. There is a fork between each plate.
 * The dish served is a kind of spaghetti which has to be eaten with two
 * forks. Each philosopher can only alternately think and eat. Moreover,
 * a philosopher can only eat their spaghetti when they have both a left and
 * right fork. Thus two forks will only be available when their two nearest
 * neighbors are thinking, not eating. After an individual philosopher
 * finishes eating, they will put down both forks. The problem is how to
 * design a regimen (a concurrent algorithm) such that any philosopher
 * will not starve; i.e., each can forever continue to alternate between
 * eating and thinking, assuming that no philosopher can know when others
 * may want to eat or think (an issue of incomplete information).
 * (From https://en.wikipedia.org/wiki/Dining_philosophers_problem.)
 * 
 * Philosophers are numbered 0 through (philCnt - 1).  Philosopher 0 must
 * acquire forks 0 ("left" fork) and 1 ("right" fork) before eating.
 * Philosopher 1 acquires forks 1 (left) and 2 (right).
 * Philosopher (philCnt - 1), the last philosopher, acquires forks
 * (philcnt - 1) (left) and 0 (right).
 * 
 * This program demonstrates the deadlock that can occur with this problem,
 * and two solutions.  Refer to the documentation for the variable 'fix'
 * below.  Read https://en.wikipedia.org/wiki/Deadlock for an introduction
 * to deadlocks and their prevention/avoidance.
 * 
 * For this lab, run the program using each of three values for fix, observe
 * the program's behavior, and carefully and clearly answer these questions:
 * 
 *     1) What causes the deadlock in this program to occur?
 *     2) How does fix 1 prevent deadlock from occurring?
 *     3) How does fix 2 prevent deadlock from occurring?
 * 
 * When the philosopher threads are deadlocked, use the red Terminate button
 * above the Console pane to terminate the program.  If you don't do this, you
 * will likely encounter problems with either Eclipse or your NoMachine
 * session.
 */


@SuppressWarnings("unused")
public class DiningPhilosophers {

	/*
	 * fix value meanings: 
	 *     0: No fixes. Deadlock will occur
	 *     1: The forks are acquired in ascending order, based upon their
	 *        index in the forks array. Deadlock is prevented
	 *     2: Only allow (philCnt - 1) philosophers to eat at any time
	 *        Deadlock is prevented.
	 */
	private final int fix = 0;

	// Number of philosophers to create.
	private final int philCnt = 3;

	// Print an "I'm alive" message each time a philosopher has completed
	// this many thinking/eating cycles.
	private final int aliveCnt = 10;

	private final ReentrantLock[] forks;

	// This counting semaphore will be used to limit the number of
	// philosophers who can simultaneously eat.
	private final Semaphore eatingLimit;

	
	public DiningPhilosophers() {
		if (fix == 2) {
			eatingLimit = new Semaphore(philCnt - 1);
		} else {
			eatingLimit = null;
		}

		forks = new ReentrantLock[philCnt];

		for (int i = 0; i < philCnt; i++) {
			forks[i] = new ReentrantLock();
		}

		for (int i = 0; i < philCnt; i++) {
			new Thread(philosopher(i)).start();
		}
	}

	
	private Runnable philosopher(int id) {
		return () -> {
			// i counts modulo aliveCnt
			int i = 0;
			// "Left" fork; acquired first.
			int left = id;
			// "Right" fork; acquired second.
			int right = (id + 1) % philCnt;

			// For fix 1, ensure the forks are acquired in ascending order.
			if ((fix == 1) && (left > right)) {
				int temp = left;
				left = right;
				right = temp;
			}

			while (true) {

				// Thinking

				// For fix 2, ensure that the number of simultaneous eating
				// philosophers is limited.
				if (fix == 2) {
					try {
						eatingLimit.acquire();
					} catch (InterruptedException e) {
					}
				}

				forks[left].lock();
				forks[right].lock();

				// Eating

				forks[left].unlock();
				forks[right].unlock();

				if (fix == 2) {
					eatingLimit.release();
				}

				// Print the "I'm alive" message periodically.
				i = (i + 1) % aliveCnt;
				if (i == 0) {
					System.out.println(id + " alive.");
				}
			}
		};
	}

	
	public static void main(String[] args) {
		new DiningPhilosophers();
	}
}
